import { CanvasSmbl, ProcessedSmbl, type CanvasElement, type Point, type Pointer, type CanvasData } from "./symbols";

export const listen = (
  el: Element | GlobalEventHandlers,
  type: string,
  listener: EventListenerOrEventListenerObject,
  once?: boolean
) => el.addEventListener(type, listener, { passive: true, once });

export const listenDel = (
  el: Element | GlobalEventHandlers,
  type: string,
  listener: EventListenerOrEventListenerObject,
  capture?: boolean
) => el?.removeEventListener(type, listener, { capture });

export const pointInCanvas = (
  canvasData: { position: { x: number; y: number }; scale: number },
  x: number,
  y: number
) => ({
  x: (x - canvasData.position.x) / canvasData.scale,
  y: (y - canvasData.position.y) / canvasData.scale,
});

export function moveScaleApplay(canvas: CanvasElement, updateScale?: { (newScale: number): void }) {
  const canvasData = canvas[CanvasSmbl]!.data;

  const gripUpdate = canvas.ownerSVGElement ? applayGrid(canvas.ownerSVGElement, canvasData) : () => {};

  function transform() {
    canvas.style.transform = `matrix(${canvasData.scale}, 0, 0, ${canvasData.scale}, ${canvasData.position.x}, ${canvasData.position.y})`;

    gripUpdate();
  }

  function scale(nextScale: number, originPoint: Point) {
    if (nextScale < 0.5 || nextScale > 4) {
      if (nextScale < 0.5) canvasData.scale = 0.5;
      if (nextScale > 4) canvasData.scale = 4;
      return;
    }
    const divis = nextScale / canvasData.scale;
    canvasData.scale = nextScale;

    canvasData.position.x = divis * (canvasData.position.x - originPoint.x) + originPoint.x;
    canvasData.position.y = divis * (canvasData.position.y - originPoint.y) + originPoint.y;

    if (updateScale) {
      updateScale(canvasData.scale);
    }

    transform();
  }

  // move, scale with fingers
  if (canvas.ownerSVGElement) {
    applayFingers(canvas.ownerSVGElement, canvasData, scale, transform);
  }

  // scale with mouse wheel
  canvas.ownerSVGElement?.addEventListener(
    "wheel",
    /** @param {WheelEvent} evt */ (evt) => {
      evt.preventDefault();
      const delta = evt.deltaY || evt.deltaX;
      const scaleStep =
        Math.abs(delta) < 50
          ? 0.05 // trackpad pitch
          : 0.25; // mouse wheel

      scale(canvasData.scale + (delta < 0 ? scaleStep : -scaleStep), evtPoint(evt));
    }
  );
  if (canvas[CanvasSmbl]) {
    canvas[CanvasSmbl].move = function (x, y) {
      canvasData.position.x = x;
      canvasData.position.y = y;
      transform();
    };
    canvas[CanvasSmbl].scale = function (newScale) {
      let bounding = canvas.ownerSVGElement;
      let width = bounding?.clientWidth ?? 0;
      let height = bounding?.clientHeight ?? 0;
      scale(newScale, { x: width / 2, y: height / 2 });
      transform();
    };
  }
}

function applayFingers(
  svg: SVGSVGElement,
  canvasData: { position: Point; scale: number },
  scaleFn: { (nextScale: number, originPoint: Point): void },
  transformFn: { (): void }
) {
  let firstPointer: Pointer | null;
  let secondPointer: Pointer | null;
  let distance: number | null;
  let center: Point | null;

  /** @param {PointerEvent} evt */
  function cancel(evt: any) {
    distance = null;
    center = null;
    if (firstPointer?.id === evt.pointerId) {
      firstPointer = null;
    }
    if (secondPointer?.id === evt.pointerId) {
      secondPointer = null;
    }

    if (!firstPointer && !secondPointer) {
      listenDel(svg, "pointermove", move);
      listenDel(svg, "pointercancel", cancel);
      listenDel(svg, "pointerup", cancel);
    }
  }

  /** @param {PointerEvent} evt */
  function move(evt: any) {
    if (evt[ProcessedSmbl]) {
      return;
    }

    if ((firstPointer && !secondPointer) || (!firstPointer && secondPointer)) {
      // move with one pointer
      canvasData.position.x = evt.clientX + ((firstPointer || secondPointer)?.shift?.x || 0);
      canvasData.position.y = evt.clientY + ((firstPointer || secondPointer)?.shift?.y || 0);
      transformFn();
      return;
    }

    if (
      !secondPointer ||
      !firstPointer ||
      (secondPointer?.id !== evt.pointerId && firstPointer?.id !== evt.pointerId)
    ) {
      return;
    }

    const distanceNew = Math.hypot(firstPointer.pos.x - secondPointer.pos.x, firstPointer.pos.y - secondPointer.pos.y);
    const centerNew = {
      x: (firstPointer.pos.x + secondPointer.pos.x) / 2,
      y: (firstPointer.pos.y + secondPointer.pos.y) / 2,
    };

    // not first move
    if (distance && center) {
      canvasData.position.x = canvasData.position.x + centerNew.x - center.x;
      canvasData.position.y = canvasData.position.y + centerNew.y - center.y;

      scaleFn((canvasData.scale / distance) * distanceNew, centerNew);
    }

    distance = distanceNew;
    center = centerNew;

    if (firstPointer.id === evt.pointerId) {
      firstPointer = evtPointer(evt, canvasData);
    }
    if (secondPointer.id === evt.pointerId) {
      secondPointer = evtPointer(evt, canvasData);
    }
  }

  listen(
    svg,
    "pointerdown",
    /** @param {PointerEvent} evt */ (evt: any) => {
      if (evt[ProcessedSmbl] || (!firstPointer && !evt.isPrimary) || (firstPointer && secondPointer)) {
        return;
      }

      svg.setPointerCapture(evt.pointerId);
      if (!firstPointer) {
        listen(svg, "pointermove", move);
        listen(svg, "pointercancel", cancel);
        listen(svg, "pointerup", cancel);
      }

      if (!firstPointer) {
        firstPointer = evtPointer(evt, canvasData);
        return;
      }
      if (!secondPointer) {
        secondPointer = evtPointer(evt, canvasData);
      }
    }
  );
}

function applayGrid(svg: SVGSVGElement, canvasData: CanvasData) {
  let curOpacity: number;
  function backImg(opacity: number) {
    if (curOpacity !== opacity) {
      curOpacity = opacity;
      svg.style.backgroundImage = `radial-gradient(rgb(73 80 87 / ${opacity}) 1px, transparent 0)`;
    }
  }

  backImg(0.7);
  svg.style.backgroundSize = `${canvasData.cell}px ${canvasData.cell}px`;

  return function () {
    const size = canvasData.cell * canvasData.scale;

    if (canvasData.scale < 0.5) {
      backImg(0);
    } else if (canvasData.scale <= 0.9) {
      backImg(0.3);
    } else {
      backImg(0.7);
    }

    svg.style.backgroundSize = `${size}px ${size}px`;
    svg.style.backgroundPosition = `${canvasData.position.x}px ${canvasData.position.y}px`;
  };
}

function evtPoint(evt: any): Point {
  return { x: evt.clientX, y: evt.clientY };
}

function evtPointer(evt: any, canvasData: { position: Point; scale: number }): Pointer {
  return {
    id: evt.pointerId,
    pos: evtPoint(evt),
    shift: {
      x: canvasData.position.x - evt.clientX,
      y: canvasData.position.y - evt.clientY,
    },
  };
}
